# -*- python -*-

from __future__ import division
from professor2.core import *
from professor2.histos import *
from professor2.errors import *
from professor2.sampling import *
import numpy as np


def mk_ipolinputs(params):
    """
    Make sorted run name and parameter name & value lists, suitable for passing to prof.Ipol

    params is a dict (actually, prefer OrderedDict) of run_names -> param_vals,
    as returned from read_rundata
    """
    runs = sorted(params.keys())
    if not runs:
        return runs, [], [[]]
    paramnames = list(params[runs[0]].keys())
    paramslist = [[params[run][pn] for pn in paramnames] for run in runs]
    return runs, paramnames, paramslist


def mk_ipolbin(rawP, rawV, rawE, xmin, xmax, CFG):
    # TODO finally learn how to use kwargs
    order    = CFG["ORDER"]
    errorder = CFG["ERR_ORDER"]
    errmode  = CFG["ERR_MODE"]
    medfilt  = CFG["MEDIAN_FILT"]

    if medfilt > 0:
        from numpy import median
        # TODO figure out what to do with x=0
        relErrs = [rawE[num]/x if x!=0 else 1 for num, x in enumerate(rawV)]
        rem = medfilt*median(relErrs)
        P, V, E = [], [], []
        for num, x in enumerate(relErrs):
            if x < rem:
                P.append(rawP[num])
                V.append(rawV[num])
                E.append(rawE[num])
        if CFG["DEBUG"]:
            print("%i/%i survive median filter %f times %f" % (len(P), len(rawP), medfilt, median(relErrs)))
    else:
        P = rawP
        V = rawV
        E = rawE
    P = np.array(P)
    V = np.array(V)
    E = np.array(E)

    import professor2 as prof
    pmin = prof.mk_minvals(P)
    pmax = prof.mk_maxvals(P)

    ## Identify nans in the inputs
    ivnans = np.isnan(V)
    vP = P[~ivnans,:]
    vV = V[~ivnans]
    ienans = np.isnan(E)
    eP = P[~ienans,:]
    eE = E[~ienans]

    ## Build the value interpolation
    try:
        if order == "auto":
            valipol = mk_autoipol(vP, vV, CFG)
        else:
            valipol = Ipol(vP, vV, int(order))
        ## Check for NaN coeffs
        import math
        if any([math.isnan(x) for x in valipol.coeffs]):
            raise NanError("NaN coefficient encountered in value ipol")
    except RuntimeError as e:
        raise IpolError(f"Could not build value ipol: {e}")

    ## Build the error interpolation(s)
    try:
        if not errmode or errmode == "none":
            erripols = None
        elif errmode == "mean":
            meanerr = sum(eE) / float(len(eE)) #histos[run].bins[binnr].err for run in runs) / float(len(runs))
            erripols = Ipol(eP, [meanerr], 0) #< const 0th order interpolation
        elif errmode == "median":
            medianerr = eE[len(eE)//2]
            erripols = Ipol(eP, [medianerr], 0) #< const 0th order interpolation
        elif errmode == "symm":
            if errorder == "auto":
                erripols = mk_autoipol(eP, eE, CFG)
            else:
                erripols = Ipol(eP, eE, int(errorder))
        elif errmode == "asymm":
            raise Exception("Error interpolation mode 'asymm' not yet supported")
        else:
            raise IpolError("Unknown error interpolation mode '%s'" % errmode)
        ## Check for NaN coeffs
        if erripols is not None:
            if any([math.isnan(x) for x in erripols.coeffs]):
                raise NanError("NaN coefficient encountered in error ipol")
    except RuntimeError as e:
        raise IpolError(f"Could not build error ipol: {e}")

    return IpolBin(xmin, xmax, valipol, erripols), pmin, pmax

def mk_autoipol(P, V, CFG):
    omin  = CFG["AUTO_OMIN"]  if CFG.has_key("AUTO_OMIN")  else 0
    omax  = CFG["AUTO_OMAX"]  if CFG.has_key("AUTO_OMAX")  else 99
    split = CFG["AUTO_SPLIT"] if CFG.has_key("AUTO_SPLIT") else 0.1
    nit   = CFG["AUTO_NIT"]   if CFG.has_key("AUTO_NIT")   else 10
    debug = CFG["DEBUG"]      if CFG.has_key("DEBUG")      else False

    # Incides of all inputs --- needed for shuffeling
    ALLI = list(range(len(P)))

    # Number of test points
    NTEST = int(split*len(ALLI))
    NTRAIN = len(ALLI)-NTEST

    # Prepare training samples
    trainings = [r for r in xrandomUniqueCombinations(ALLI, len(ALLI)-NTEST, nit)]
    # Prepare test samples
    tests = [[a for a in ALLI if not a in t] for t in trainings]

    # Dimension of parameter space
    DIM = len(P[0])

    # Get possible orders
    ORDERS = []
    o_temp = omin
    import professor2 as prof
    while True:
        n_temp = prof.calcnumCoeffs(DIM, o_temp)
        if n_temp > NTRAIN or o_temp > omax:
            break
        ORDERS.append(o_temp)
        o_temp+=1

    if len(ORDERS)==0:
        raise Exception("Not enough data to do anything")

    residuals, meanresiduals, meanresidualstimesncoeff = {}, {}, {}
    for o in ORDERS:
        residuals[o] = []
        # Iterate through training "samples"
        for num, train in enumerate(trainings):
            # Calculate ipol for this run combination
            thisP = [P[x] for x in train]
            thisV = [V[x] for x in train]
            thisI = Ipol(thisP, thisV, o)
            # Get the residuals for all test points
            thisRes = [(thisI.val(P[x]) - V[x])**2 for x in range(len(tests[num]))]
            residuals[o].extend(thisRes)

    from numpy import mean
    for k, v in residuals.items():
        meanresiduals[k] = mean(v)
        meanresidualstimesncoeff[k] = mean(v) * prof.calcnumCoeffs(DIM, k)
    #winner=min(meanresiduals, key=meanresiduals.get)
    winner=min(meanresidualstimesncoeff, key=meanresidualstimesncoeff.get)
    if debug:
        print("Residual summary:")
        print("Choose order %i"%winner)
        for k, v in meanresiduals.items():
            print("%i: %e times %i coeffs = %e"%(k, v, numCoeffs(DIM, k), meanresidualstimesncoeff[k]))
    return Ipol(P, V, winner)


## Keep this for backward compatibility
def mk_ipolhisto(histos, runs, paramslist, order, errmode=None, errorder=None):
    """\
    Make a prof.IpolHisto from a dict of prof.DataHistos and the corresponding
    runs and params lists, at the given polynomial order.

    If errs is non-null, the data histo errors will also be interpolated.

    If errmode is None or 'none', uncertainties will not be parameterised and
    will return 0 if queried; 'mean' and 'median' will use fixed values derived
    from the anchor points; 'symm' will parameterise the average of the + and -
    errors of each bin at the polynomial order given by errorder. If errorder is
    None, the same order as for the value parameterisation will be used.

    Parameter range scaling will be applied, so a DoParamScaling=true flag will
    need to be written to the metadata when persisting the resulting IpolHisto.

    """
    if errmode is None:
        errmode = "none"
    if errorder is None:
        errorder = order
    #
    nbins = len(iter(histos.values()).next().bins)
    ibins = []
    for n in range(nbins):
        ## Check that the bin edges are consistent and extract their values
        # TODO: move bin edge consistency checking into the Histo base class
        xmax = list(histos.values())[0].bins[n].xmax
        xmin = list(histos.values())[0].bins[n].xmin
        vals = [histos[run].bins[n].val for run in runs]
        errs = [histos[run].bins[n].err for run in runs]
        try:
            ibins.append(mk_ipolbin(paramslist, vals, errs, xmin, xmax, order, errmode, errorder))
        except NanError as ne:
            print(ne, "in bin %i of %s" % (n, list(histos.values())[0].path))
    return Histo(ibins, list(histos.values())[0].path)


# https://stackoverflow.com/questions/2130016/splitting-a-list-of-into-n-parts-of-approximately-equal-length
def chunkIt(seq, num):
    avg = len(seq) / float(num)
    out = []
    last = 0.0
    while last < len(seq):
        out.append(seq[int(last):int(last + avg)])
        last += avg
    return out


def mkStandardIpols(HISTOS, HNAMES, RUNS, PARAMSLIST, CFG, nchunks=10, referenceIpolSet=None, quiet=False):

    """
    the referenceIpolSet allows to make structurally identical ipols
    """
    BNAMES = []
    for hn in HNAMES:
        histos = HISTOS[hn]
        nbins = list(histos.values())[0].nbins
        for n in range(nbins):
            BNAMES.append([hn, n])
            if referenceIpolSet is not None:
                b_order_val = referenceIpolSet[hn].bins[n].ival.order
                b_order_err = referenceIpolSet[hn].bins[n].ierrs.order
                BNAMES[-1].append(b_order_val)
                BNAMES[-1].append(b_order_err)

    NBINS = len(BNAMES)

    MSGEVERY = int(NBINS/100.) if NBINS > 100 else 1;

    import sys, zlib
    import professor2 as prof
    def worker(q, rdict, counter):
        "Function to make bin ipols and store ipol persistency strings for each histo"
        import sys
        while True:
            if q.empty():
                break
            try:
                temp = q.get(False)
            except Exception as e:
                print("Oops: {}".format(e))
                break

            hn = temp[0]
            histos = HISTOS[hn]
            n = temp[1]
            xmax = list(histos.values())[0].bins[n].xmax
            xmin = list(histos.values())[0].bins[n].xmin


            from math import log,isnan
            if CFG["LOGY"]:
                vals = [log(histos[run].bins[n].val) if histos[run].bins[n].val > 0 else  0 for run in RUNS if not (isnan(histos[run].bins[n].val) or isnan(histos[run].bins[n].err))]
                errs = [log(histos[run].bins[n].err) if histos[run].bins[n].err > 0 else -1 for run in RUNS if not (isnan(histos[run].bins[n].val) or isnan(histos[run].bins[n].err))]
            else:
                vals = [histos[run].bins[n].val for run in RUNS if not (isnan(histos[run].bins[n].val) or isnan(histos[run].bins[n].err))]
                errs = [histos[run].bins[n].err for run in RUNS if not (isnan(histos[run].bins[n].val) or isnan(histos[run].bins[n].err))]

            if len(RUNS)-len(vals) > 0:
                print(f"{len(RUNS)-len(vals)}/{len(RUNS)} runs are ignored due to NANs in values or errors for ipol of {hn}#{n}")
            #ib, pmin, pmax = prof.mk_ipolbin(pars, vals, errs, xmin, xmax, CFG)
            try:
                if referenceIpolSet is not None:
                    CFG["ORDER"] = temp[2]
                    CFG["ERR_ORDER"] = temp[3]
                    ib, pmin, pmax = prof.mk_ipolbin(PARAMSLIST, vals, errs, xmin, xmax, CFG)
                else:
                    ib, pmin, pmax = prof.mk_ipolbin(PARAMSLIST, vals, errs, xmin, xmax, CFG)

                s = ""
                s += "%s#%d %.5e %.5e\n" % (hn, n, ib.xmin, ib.xmax)
                s += "  " + ib.ival.toString("val")
                for v in pmin:
                    s += " %.5e" % v
                for v in pmax:
                    s += " %.5e" % v
                s += "\n"
                if ib.ierrs:
                    s += "  " + ib.ierrs.toString("err")
                    for v in pmin:
                        s += " %.5e" % v
                    for v in pmax:
                        s += " %.5e" % v
                    s += "\n"
                rdict.put( [hn,n, zlib.compress(s.encode("utf-8"), 9)])
                del s
                del ib #< pro-actively clear up memory

            except NanError as ne:
                print(ne, "in bin %i of %s" % (n, list(histos.values())[0].path))
            except IpolError as ie:
                print(ie, "in bin %i of %s" % (n, list(histos.values())[0].path))

            del histos
            counter.value += 1
            if counter.value == MSGEVERY and not quiet:
                counter.value = 0
                sys.stderr.write('\rProgress: {current}/{total}\r'.format(current=rdict.qsize(), total=NBINS))
            q.task_done()
        return

    # TODO: Printing and multiprocessing should happen under script control
    if not CFG["QUIET"]:
        print("\nParametrising %i objects...\n" % len(BNAMES))
    import time, multiprocessing
    rDict = {}
    time1 = time.time()

    from multiprocessing import Manager, Process
    manager = Manager()

    # This for the status --- modulus is too expensive
    ndone = manager.Value('i', 0)
    ## A shared memory object is required for coefficient retrieval
    r = manager.Queue()

    # # For testing with IPython embed --- leave the following 4 lines in please
    # q = manager.Queue()
    # for chunk in chunkIt(BNAMES, nchunks): # The chunking is necessary as the memory blows up otherwise
        # for x in chunk:q.put(x)
        # worker(q, r , ndone)

    # from IPython import embed
    # embed()

    for chunk in chunkIt(BNAMES, nchunks): # The chunking is necessary as the memory blows up otherwise

        ## The job queue
        q = manager.Queue()

        ## Fire away
        workers = [Process(target=worker, args=(q, r, ndone)) for i in range(CFG["MULTI"])]
        [q.put(x)  for x in  chunk]
        [x.start() for x in workers]

        [x.join() for x in  workers]
        [x.terminate() for x in  workers]
        # map(lambda x:q.put(x), chunk)
        # map(lambda x:x.start(), workers)

        # map(lambda x:x.join(),  workers)
        # map(lambda x:x.terminate(),  workers)

    ## Timing
    while not r.empty():
        a, b, c = r.get()
        rDict[(a,b)] =c

    time2 = time.time()
    if not CFG["QUIET"]:
        print(('\n\nParametrisation took %0.2fs' % ((time2-time1))))

    return rDict


def writeIpol(fname, ipolDict, params, runs=[], summary="", runsdir=""):
    PARAMNAMES = params[0]
    PARAMSLIST = params[1]

    import os, tempfile, zlib
    if fname == "temp":
        f = tempfile.NamedTemporaryFile(delete=False)
    else:
        f = open(fname, "w")

    import professor2 as prof
    f.write("Summary: %s\n" % summary)
    f.write("DataDir: %s\n" % os.path.abspath(runsdir))
    f.write("ProfVersion: %s\n" % prof.version())
    f.write("Date: %s\n" % prof.mk_timestamp())
    f.write("DataFormat: binned 3\n") # This tells the reader how to treat the coefficients that follow
    # Format and write out parameter names
    pstring = "ParamNames:"
    for p in PARAMNAMES:
        pstring += " %s" % p
    f.write(pstring + "\n")
    # Dimension (consistency check)
    f.write("Dimension: %i\n" % len(PARAMNAMES))
    # Interpolation validity (hypercube edges)
    # TODO a lot of this is obsolete in the format 3
    minstring = "MinParamVals:"
    for v in prof.mk_minvals(PARAMSLIST):
        minstring += " %f" % v
    f.write(minstring + "\n")
    maxstring = "MaxParamVals:"
    for v in prof.mk_maxvals(PARAMSLIST):
        maxstring += " %f" % v
    f.write(maxstring + "\n")
    f.write("DoParamScaling: 1\n")
    # Number of inputs per bin
    f.write("NumInputs: %i\n" % len(PARAMSLIST))
    s_runs = "Runs:"
    for r in runs:
        s_runs +=" %s"%r
    f.write("%s\n"%s_runs)
    f.write("---\n")

    ## Write out numerical data for all interpolations
    s = ""
    HNAMES=sorted(list(set([x[0] for x in list(ipolDict.keys())])))
    for hn in sorted(HNAMES):
        thisbins=sorted(filter(lambda x: x[0]==hn, list(ipolDict.keys())))
        for ipolstring in [ipolDict[x] for x in thisbins]:
            s+=zlib.decompress(ipolstring).decode("utf-8")

    f.write(s)
    f.close()
    if not fname == "temp":
        print("\nOutput written to %s" % fname)
    else:
        return f




def getBox(P, masterbox, mastercenter, debug=False):
    """
    Helper function that finds the most appropriate box for a parameter point P
    """
    import professor2 as prof
    boxdict=None
    for box, bdict in masterbox.items():
        if prof.pInBOX(P, box, debug):
            boxdict = bdict
            break
    if boxdict is None:
        distances={}
        for c in list(mastercenter.keys()):
            distances[prof.pBoxDistance(P, c)] = c
        winner = min(distances.keys())
        boxdict = mastercenter[distances[winner]]
    return boxdict
